@@ -38,13 +38,14 @@ pub struct BucketMapHolderStats {
     pub count_in_mem: AtomicUsize,
     pub per_bucket_count: Vec<AtomicUsize>,
     pub flush_entries_updated_on_disk: AtomicU64,
-    pub flush_entries_removed_from_mem: AtomicU64,
+    pub flush_entries_evicted_from_mem: AtomicU64,
     pub active_threads: AtomicU64,
     pub get_range_us: AtomicU64,
     last_age: AtomicU8,
     last_ages_flushed: AtomicU64,
-    pub flush_scan_update_us: AtomicU64,
-    pub flush_remove_us: AtomicU64,
+    pub flush_scan_us: AtomicU64,
+    pub flush_update_us: AtomicU64,
+    pub flush_evict_us: AtomicU64,
     pub flush_grow_us: AtomicU64,
     last_was_startup: AtomicBool,
     last_time: AtomicInterval,
@@ -153,6 +154,16 @@ impl BucketMapHolderStats {
         }
     }
 
+    /// This is an estimate of the # of items in mem that are awaiting flushing to disk.
+    /// returns (# items in mem) - (# items we intend to hold in mem for performance heuristics)
+    /// The result is also an estimate because 'held_in_mem' is based on a stat that is swapped out when stats are reported.
+    pub fn get_remaining_items_to_flush_estimate(&self) -> usize {
+        let in_mem = self.count_in_mem.load(Ordering::Relaxed) as u64;
+        let held_in_mem = self.held_in_mem_slot_list_cached.load(Ordering::Relaxed)
+            + self.held_in_mem_slot_list_len.load(Ordering::Relaxed);
+        in_mem.saturating_sub(held_in_mem) as usize
+    }
+
     pub fn report_stats<T: IndexValue>(&self, storage: &BucketMapHolder<T>) {
         let elapsed_ms = self.last_time.elapsed_ms();
         if elapsed_ms < STATS_INTERVAL_MS {
@@ -329,8 +340,13 @@ impl BucketMapHolderStats {
                 ("keys", self.keys.swap(0, Ordering::Relaxed), i64),
                 ("ms_per_age", ms_per_age, i64),
                 (
-                    "flush_scan_update_us",
-                    self.flush_scan_update_us.swap(0, Ordering::Relaxed),
+                    "flush_scan_us",
+                    self.flush_scan_us.swap(0, Ordering::Relaxed),
+                    i64
+                ),
+                (
+                    "flush_update_us",
+                    self.flush_update_us.swap(0, Ordering::Relaxed),
                     i64
                 ),
                 (
@@ -339,8 +355,8 @@ impl BucketMapHolderStats {
                     i64
                 ),
                 (
-                    "flush_remove_us",
-                    self.flush_remove_us.swap(0, Ordering::Relaxed),
+                    "flush_evict_us",
+                    self.flush_evict_us.swap(0, Ordering::Relaxed),
                     i64
                 ),
                 (
@@ -432,8 +448,8 @@ impl BucketMapHolderStats {
                     i64
                 ),
                 (
-                    "flush_entries_removed_from_mem",
-                    self.flush_entries_removed_from_mem
+                    "flush_entries_evicted_from_mem",
+                    self.flush_entries_evicted_from_mem
                         .swap(0, Ordering::Relaxed),
                     i64
                 ),
